// standard includes
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

// class declaration
#include "curl.hpp"

#include <errno.h>

using namespace std;

pthread_mutex_t MyMultiCURL::mutex = PTHREAD_MUTEX_INITIALIZER;
size_t MyMultiCURL::size = (!getenv("MY_CURL_BUFFER_SIZE")) ? CURL_MAX_WRITE_SIZE : atoi(getenv("MY_CURL_BUFFER_SIZE"));

MyMultiCURL::MyMultiCURL(CURLM *multicurl) : multicurl(multicurl)
{
  /* we can optionally limit the total amount of connections this multi handle uses */
  curl_multi_setopt(multicurl, CURLMOPT_MAXCONNECTS, (long)MAX_CONNECTIONS);
  curl_multi_setopt(multicurl, CURLMOPT_PIPELINING, (long)1);
  
  /* setup the generic multi interface options we want */
  curl_multi_setopt(multicurl, CURLMOPT_SOCKETFUNCTION, multicurl_socket_cb);
  curl_multi_setopt(multicurl, CURLMOPT_TIMERFUNCTION, multicurl_timer_cb);

  eventbase = event_base_new();  //event_init();
  cout <<  "event base method used is "<< event_base_get_method(eventbase) << endl;
}

MyMultiCURL::~MyMultiCURL(void) 
{ 
  cout <<  "cleaning up curl handle..." << endl;
  
  if (eventbase)
    event_base_free(eventbase);
  
  if (multicurl)
    curl_multi_cleanup(multicurl);
}

int MyMultiCURL::new_conn(const char *url, char*& buffer, int current_uri_pos, CURL *easy)
{
  CURLUserData *data = (CURLUserData *)malloc(1*sizeof(CURLUserData));
 
  memset(data, 0, sizeof(CURLUserData));
  // set user data parameters
  data->buffer = buffer;
  data->cup = current_uri_pos;
  data->status = -1; // this is used only in case of error in writeData
  data->error[0] = '\0';

  if (!easy) {
    data->easy = curl_easy_init();     // for each file
    curl_easy_setopt(data->easy, CURLOPT_PROTOCOLS, CURLPROTO_HTTP); // for each file
    curl_easy_setopt(data->easy, CURLOPT_REDIR_PROTOCOLS, CURLPROTO_HTTP); // for each file
    curl_easy_setopt(data->easy, CURLOPT_NOPROGRESS, 1); // for each file
    curl_easy_setopt(data->easy, CURLOPT_HEADER, 0); // for each file
    curl_easy_setopt(data->easy, CURLOPT_HTTPHEADER, NULL);
    curl_easy_setopt(data->easy, CURLOPT_USE_SSL, CURLUSESSL_NONE);
    //curl_easy_setopt(data->easy, CURLOPT_PORT, 85); // for each file
    //curl_easy_setopt(data->easy, CURLOPT_DNS_CACHE_TIMEOUT, 0);
    curl_easy_setopt(data->easy, CURLOPT_HTTPGET, 1); //for each file
    curl_easy_setopt(data->easy, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1); //for each file
    curl_easy_setopt(data->easy, CURLOPT_FRESH_CONNECT, 0); //for each file. Tell curl to use any cached connection
    //curl_easy_setopt(data->easy, CURLOPT_FORBID_REUSE, 0); //for each file. Tell curl not to close connection immediately after the transfer
    //curl_easy_setopt(data->easy, CURLOPT_TCP_NODELAY, 1);
    //curl_easy_setopt(data->easy, CURLOPT_VERBOSE, 1L);
    curl_easy_setopt(data->easy, CURLOPT_NOSIGNAL, 1L);
    curl_easy_setopt(data->easy, CURLOPT_DNS_CACHE_TIMEOUT, 0);
  }
  else  {
    data->easy = easy;
  }

  curl_easy_setopt(data->easy, CURLOPT_BUFFERSIZE, MyMultiCURL::size); // for each file
  curl_easy_setopt(data->easy, CURLOPT_WRITEFUNCTION, MyMultiCURL::writeData); // for each file
  curl_easy_setopt(data->easy, CURLOPT_PRIVATE, (void *)data);
  curl_easy_setopt(data->easy, CURLOPT_URL, url); // for each file  
  curl_easy_setopt(data->easy, CURLOPT_WRITEDATA, (void *) data); // for each file
  curl_easy_setopt(data->easy, CURLOPT_ERRORBUFFER, data->error); // for each file
  curl_multi_add_handle(multicurl, data->easy);            // for each file

  return OK;
}

void MyMultiCURL::setsock(CURLSocketData *sockdata, int action, curl_socket_t s, CURLGlobalData *g)
{
  //tell libevent that we are interested in READ (server response) and write (request to server)
  int event_action = (action & CURL_POLL_IN?EV_READ:0)|(action & CURL_POLL_OUT?EV_WRITE:0)|EV_PERSIST;

  sockdata->sockfd = s;
  if (sockdata->evset)         
    event_free(sockdata->ev);   //event_free(&sockdata->ev); 
  //event_set(&sockdata->ev, sockdata->sockfd, event_action, libevent_event_cb, g);
  sockdata->ev = event_new(g->eventbase, sockdata->sockfd, event_action, libevent_event_cb, g);
  event_add(sockdata->ev, NULL);  //event_add(&sockdata->ev, NULL); 
  sockdata->evset=1; //set this to 1 after libevent is set to monitor.
  return;
}

int MyMultiCURL::multicurl_socket_cb(CURL *easy, curl_socket_t curlsocket, int action, void *userp, void *socketp)
{
  CURLGlobalData *g = (CURLGlobalData *)userp;
  CURLSocketData *usersockdata = (CURLSocketData *)socketp;

  if (action == CURL_POLL_REMOVE) {
    if (usersockdata){
      if (usersockdata->evset)
          event_free(usersockdata->ev); //event_del(&usersockdata->ev);
      usersockdata->evset = 0;
      free(usersockdata); usersockdata = NULL;
    }
  }
  else {
    if (!usersockdata) {
      if (NULL == (usersockdata = (CURLSocketData *)malloc(sizeof(CURLSocketData)) ) )
      {
        //some problem occurred with malloc. so, log the message.
        cout <<  "Unable to create CURLSocketData..." << endl;
        return -1; //TODO: check the return values
      }
      memset(usersockdata, '\0', sizeof(CURLSocketData));
      setsock(usersockdata, action, curlsocket, g);
      curl_multi_assign(g->multicurl, curlsocket, usersockdata);
    }
    else {
      setsock(usersockdata, action, curlsocket, g);
    }
  }
  return 0; //multi socket call back must return 0 on success (per tutorial)
}

void MyMultiCURL::libevent_event_cb(int fd, short kind, void *userp)
{
  CURLGlobalData *g = (CURLGlobalData *)userp;
  CURLMcode rc;
  pthread_t thread;

  int action = (kind & EV_READ ? CURL_CSELECT_IN : 0) | (kind & EV_WRITE ? CURL_CSELECT_OUT : 0);
  
  rc = curl_multi_socket_action(g->multicurl, fd, action, &g->still_running);

  check_multi_info(g);
  
  if (g->still_running <= 0 ) {
    //cout <<  "last transfer done, kill timeout\n";
    if (g->endstatus == OK)
        stopAndPrintTime("HTTP transfer time"); //Stop timer and print out time taken for HTTP downloading/transfers
    if (evtimer_pending(g->timer_event, NULL)) {
      cout <<  "last transfer done, deleting timeout\n";
      evtimer_del(g->timer_event);
    }
  }
}

int MyMultiCURL::multicurl_timer_cb(CURLM *multi, long timeout_ms, void *userp)
{
  CURLGlobalData *g = (CURLGlobalData *)userp;
  struct timeval timeout;
  
  timeout.tv_sec = timeout_ms/1000;
  timeout.tv_usec = (timeout_ms%1000)*1000;
  fprintf(stdout, "multicurl_timer_cb: Setting timeout to %ld ms\n", timeout_ms);
  evtimer_add(g->timer_event, &timeout);
  return 0;
}
  
void MyMultiCURL::libevent_timer_cb(int fd, short kind, void *userp)
{
  CURLGlobalData *g = (CURLGlobalData *)userp;
  CURLMcode rc;
  
  rc = curl_multi_socket_action(g->multicurl, CURL_SOCKET_TIMEOUT, 0, &g->still_running);
  
  check_multi_info(g);
}

/* Check for completed transfers, and remove/resest their easy handles */
void MyMultiCURL::check_multi_info(CURLGlobalData *g)
{ 
  CURLMsg *msg;
  int msgs_left;
  
  //fprintf(stdout, "REMAINING: %d\n", g->still_running);
  int easy_status = OK;  
  while ((msg = curl_multi_info_read(g->multicurl, &msgs_left))) {
      char *eff_url;
      CURLUserData *data;
      CURL *easycurl = NULL;
      if (msg->msg == CURLMSG_DONE) {
        easycurl = msg->easy_handle;
        CURLcode error_code = (CURLcode) 0;
         
        //
        // Handle error_code and other things
        // 

        curl_multi_remove_handle(g->multicurl, easycurl);  
        int c =  g->next_uri_toprocess, n_uris = g->n_uris;
        if (c < n_uris) {
          curl_easy_reset(easycurl); //curl_easy_cleanup(easycurl);  
          g->handle->new_conn(g->uris[c].address, g->buffers[c], c, easycurl); //Create a socket/transfer and add to multi stack
          g->next_uri_toprocess++;
        }
        else { //we are done with sending requests to HTTP IIS server
          curl_easy_cleanup(easycurl);
        }  
      }
      g->endstatus = OK;        
    } //while ((msg = curl_multi_info_read(multicurl, &Q)))
}


size_t MyMultiCURL::writeData(void *buffer, size_t size, size_t n, void *user_data)
{
  CURLUserData *data = (CURLUserData *) user_data;
  size_t        total = n * size;
  
  //
  // HANDLE data returned via user_data
  //
  
  return total;
}

int MyMultiCURL::batchGetData(int n_uris, URI *uris, char **buffers)
{
  memset(&g, 0, sizeof(CURLGlobalData));
  
  curlglobaldata.multicurl = multicurl;
  curlglobaldata.uris = uris;
  curlglobaldata.buffers = buffers;
  curlglobaldata.eventbase = eventbase;
  curlglobaldata.endstatus = -1;  //holds batch status
  curlglobaldata.handle = this;
  curlglobaldata.n_uris = n_uris;
  curlglobaldata.next_uri_toprocess = 0;

  curlglobaldata.timer_event = event_new(eventbase, -1, 0, libevent_timer_cb, &curlglobaldata);

  curl_multi_setopt(curlglobaldata.multicurl, CURLMOPT_SOCKETDATA, &curlglobaldata);
  curl_multi_setopt(curlglobaldata.multicurl, CURLMOPT_TIMERDATA, &curlglobaldata);

  //For now, assume we have more than 10 uris to process 
  int c = 0;
  for(;c<10;c++) {
    if (!uris[c].address || !uris[c].address[0])
      continue;
    new_conn(uris[c].address, buffers[c], c); //Create a socket/transfer and add to multi stack
  }

  curlglobaldata.next_uri_toprocess = c;
  
  startTimer();  //Start the timer
  event_base_loop(eventbase, 0); 
  
  if (curlglobaldata.timer_event) event_free(curlglobaldata.timer_event);

  return curlglobaldata.endstatus;
}

int batchGetData(int n_uris, URI *uris, char **buffers)
{
  int status;
  MyMultiCURL *handle =  NULL; 

  // sanity: check arguments
  if (n_uris <= 0 || !uris || !buffers) {
    cout << "invalid arguments..." << endl;
    return -1;
  }

  pthread_mutex_lock(&mutex);
  CURLM *multicurl;
    
  if (!(multicurl = curl_multi_init())) {
      cout <<  "error initializing curl handle..." << endl;
      pthread_mutex_unlock(&mutex);
      return -1;
  }
  handle = new MyMultiCURL(multicurl, true);  
  if (!handle) {
    cout <<  "out of address space..." << endl;
    pthread_mutex_unlock(&mutex);
    return -1;
  }
    
  cout <<  "added new curl handle..." << endl;
  
  if ((status = MyMultiCURL::grabHandle(handle)) != OK)  {
    // could not get a libcurl handle (error was printed out accordingly)
    free(handle);
    return status;
  }

  status = handle->batchGetData(n_uris, uris, buffers);
  free(handle);
  return status;
}

int external_function(param1, param2, ...)
{
  curl_global_init(CURL_GLOBAL_ALL);
  // Method which creates
  for(i = 1;i<=10<i++) {
    //code which creates threads and invokes batchGetData
  }
  curl_global_cleanup();
}